@@ -1368,6 +1368,8 @@ void check_installed_files_from_package(Options *op, Package *p)
         ui_status_update(op, percent, p->entries[i].dst);
         
         if (p->entries[i].flags & FILE_TYPE_SYMLINK) {
+            /* Don't bother checking FILE_TYPE_NEWSYMs because we may not have
+             * installed them. */
             if (!check_symlink(op, p->entries[i].target,
                                p->entries[i].dst,
                                p->description)) {
@@ -2310,11 +2312,24 @@ int check_for_nvidia_graphics_devices(Options *op, Package *p)
             int found_legacy_device = FALSE;
             for (i = 0; i < sizeof(LegacyList) / sizeof(LEGACY_INFO); i++) {
                 if (dev->device_id == LegacyList[i].uiDevId) {
+                    int j, nstrings;
+                    const char *branch_string = "";
+                    nstrings = sizeof(LegacyStrings) / sizeof(LEGACY_STRINGS);
+                    for (j = 0; j < nstrings; j++) {
+                        if (LegacyStrings[j].branch == LegacyList[i].branch) {
+                            branch_string = LegacyStrings[j].description;
+                            break;
+                        }
+                    }
+
                     ui_warn(op, "The NVIDIA %s GPU installed in this system is supported "
-                            "through the NVIDIA legacy Linux graphics drivers.  Please "
+                            "through the NVIDIA %s legacy Linux graphics drivers.  Please "
                             "visit http://www.nvidia.com/object/unix.html for more "
                             "information.  The %s NVIDIA Linux graphics driver will "
-                            "ignore this GPU.", LegacyList[i].AdapterString, p->version);
+                            "ignore this GPU.",
+                            LegacyList[i].AdapterString,
+                            branch_string,
+                            p->version);
                     found_legacy_device = TRUE;
                 }
             }
@@ -2540,6 +2555,224 @@ done:
 }
 
 
+
+/*
+ * Determine if the nouveau driver is currently in use.  We do the
+ * equivalent of:
+ *
+ *   ls -l /sys/bus/pci/devices/ /driver | grep nouveau
+ *
+ * The directory structure under /sys/bus/pci/devices/ should contain
+ * a directory for each PCI device, and for those devices with a
+ * kernel driver there will be a "driver" symlink.
+ *
+ * This appears to be consistent with how libpciaccess works.
+ *
+ * Returns TRUE if nouveau is found; returns FALSE if not.
+ */
+
+#define SYSFS_DEVICES_PATH "/sys/bus/pci/devices"
+
+static int nouveau_is_present(void)
+{
+    DIR *dir;
+    struct dirent * ent;
+    int found = FALSE;
+
+    dir = opendir(SYSFS_DEVICES_PATH);
+
+    if (!dir) {
+        return FALSE;
+    }
+
+    while ((ent = readdir(dir)) != NULL) {
+
+        char driver_path[PATH_MAX];
+        char symlink_target[PATH_MAX];
+        char *name;
+        ssize_t ret;
+
+        if ((strcmp(ent->d_name, ".") == 0) ||
+            (strcmp(ent->d_name, "..") == 0)) {
+            continue;
+        }
+
+        snprintf(driver_path, PATH_MAX,
+                 SYSFS_DEVICES_PATH "/%s/driver", ent->d_name);
+
+        driver_path[PATH_MAX - 1] = '\0';
+
+        ret = readlink(driver_path, symlink_target, PATH_MAX);
+        if (ret < 0) {
+            continue;
+        }
+
+        /* readlink(3) does not nul-terminate its returned string */
+
+        ret = NV_MIN(ret, PATH_MAX - 1);
+
+        symlink_target[ret] = '\0';
+
+        name = basename(symlink_target);
+
+        if (strcmp(name, "nouveau") == 0) {
+            found = TRUE;
+            break;
+        }
+    }
+
+    closedir(dir);
+
+    return found;
+}
+
+
+
+/*
+ * Write a modprobe configuration fragment to disable loading of
+ * nouveau:
+ *
+ *  if [ -d /etc/modprobe.d ]; then
+ *      name=/etc/modprobe.d/nvidia-installer-nouveau-blacklist
+ *      echo "# generated by nvidia-installer" > $name
+ *      echo "blacklist nouveau" >> $name
+ *      echo "options nouveau modeset=0" >> $name
+ *  fi
+ *
+ * Returns TRUE if successful; returns FALSE if there was a failure.
+ */
+
+#define ETC_MODPROBE_D "/etc/modprobe.d"
+#define ETC_MODPROBE_D_FILE \
+  ETC_MODPROBE_D "/nvidia-installer-disable-nouveau.conf"
+
+/*
+ * this checksum is the result of compute_crc() for the file contents
+ * written in blacklist_nouveau()
+ */
+
+#define ETC_MODPROBE_D_FILE_CKSUM 3728279991U
+
+static int blacklist_nouveau(void)
+{
+    int ret;
+    struct stat stat_buf;
+    FILE *file;
+
+    ret = stat(ETC_MODPROBE_D, &stat_buf);
+
+    if (ret != 0) {
+        return FALSE;
+    }
+
+    file = fopen(ETC_MODPROBE_D_FILE, "w+");
+
+    if (!file) {
+        return FALSE;
+    }
+
+    fprintf(file, "# generated by nvidia-installer\n");
+    fprintf(file, "blacklist nouveau\n");
+    fprintf(file, "options nouveau modeset=0\n");
+
+    ret = fclose(file);
+
+    if (ret != 0) {
+        return FALSE;
+    }
+
+    return TRUE;
+}
+
+
+
+/*
+ * Check if the nouveau blacklist file is already present with the
+ * contents that we expect
+ */
+
+static int nouveau_blacklist_file_is_present(Options *op)
+{
+    if (access(ETC_MODPROBE_D_FILE, R_OK) != 0) {
+        return FALSE;
+    }
+
+    if (compute_crc(op, ETC_MODPROBE_D_FILE) != ETC_MODPROBE_D_FILE_CKSUM) {
+        return FALSE;
+    }
+
+    return TRUE;
+}
+
+
+
+/*
+ * Check if the nouveau kernel driver is in use.  If it is, provide an
+ * appropriate error message and offer to try to disable nouveau.
+ *
+ * Returns FALSE if the nouveau kernel driver is in use (cause
+ * installation to abort); returns TRUE if the nouveau driver is not
+ * in use.
+ */
+
+int check_for_nouveau(Options *op)
+{
+    int ret;
+
+#define NOUVEAU_POINTER_MESSAGE                                         \
+    "Please consult the NVIDIA driver README and your Linux "           \
+        "distribution's documentation for details on how to correctly " \
+        "disable the Nouveau kernel driver."
+
+    if (op->no_nouveau_check) return TRUE;
+
+    if (!nouveau_is_present()) return TRUE;
+
+    ui_error(op, "The Nouveau kernel driver is currently in use "
+             "by your system.  This driver is incompatible with the NVIDIA "
+             "driver, and must be disabled before proceeding.  "
+             NOUVEAU_POINTER_MESSAGE);
+
+    if (nouveau_blacklist_file_is_present(op)) {
+        ui_warn(op, "The modprobe configuration file to disable Nouveau, "
+                ETC_MODPROBE_D_FILE ", is already present.  Please be "
+                "sure you have rebooted your system since that file was "
+                "written.  If you have rebooted, then Nouveau may be enabled "
+                "for other reasons, such as being included in the system "
+                "initial ramdisk or in your X configuration file.  "
+                NOUVEAU_POINTER_MESSAGE);
+        return FALSE;
+    }
+
+    ret = ui_yes_no(op, FALSE, "For some distributions, Nouveau can be "
+                    "disabled by adding a file in the modprobe configuration "
+                    "directory.  Would you like nvidia-installer to attempt "
+                    "to create this modprobe file for you?");
+
+    if (ret) {
+        ret = blacklist_nouveau();
+
+        if (ret) {
+            ui_message(op, "The modprobe configuration file to disable "
+                       "Nouveau, " ETC_MODPROBE_D_FILE ", has been written.  "
+                       "For some distributions, this may be sufficient to "
+                       "disable Nouveau; other distributions may require "
+                       "modification of the initial ramdisk.  Please reboot "
+                       "your system and attempt NVIDIA driver installation "
+                       "again.  Note if you later wish to reenable Nouveau, "
+                       "you will need to delete the file " 
+                       ETC_MODPROBE_D_FILE ".");
+        } else {
+            ui_warn(op, "Unable to alter the nouveau modprobe configuration.  "
+                    NOUVEAU_POINTER_MESSAGE);
+        }
+    }
+
+    return FALSE;
+}
+
+
+
 /*
  * nv_format_text_rows() - this function breaks the given string str
  * into some number of rows, where each row is not longer than the
